# -*- coding: utf-8 -*-
# =============================================================================
#     filter_functions
#     Copyright (C) 2019 Quantum Technology Group, RWTH Aachen University
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#     Contact email: tobias.hangleiter@rwth-aachen.de
# =============================================================================
"""
This module defines the Basis class, a subclass of NumPy's ndarray, to
represent operator bases.

Classes
-------
:class:`Basis`
    The operator basis as an array of  shape (d**2, d, d) with d the dimension
    of the Hilbert space

Functions
---------
:func:`normalize`
    Function to normalize a ``Basis`` instance
:func:`expand`
    Function to expand an array of operators in a given basis
:func:`ggm_expand`
    Fast function to expand an array of operators in a Generalized Gell-Mann
    basis

"""

from typing import Optional, Sequence, Union
from warnings import warn

import numpy as np
import opt_einsum as oe
from numpy.core import ndarray
from numpy.linalg import norm
from qutip import Qobj
from scipy.linalg import null_space
from sparse import COO

from .util import P_np, remove_float_errors, tensor

__all__ = ['Basis', 'expand', 'ggm_expand', 'normalize']


class Basis(ndarray):
    r"""
    Class for operator bases. There are several ways to instantiate a Basis
    object:

        - by just calling this constructor with a (possibly incomplete) array
          of basis matrices. In this case, it is attempted to construct from
          the input a complete basis with the following properties that retains
          all original (input) elements:

              - hermitian
              - orthonormal
              - [traceless] (can be controlled by a flag)

        - by calling one of the classes alternative constructors
          (classmethods) for predefined bases:

              - :meth:`pauli`: Pauli operator basis
              - :meth:`ggm`: Generalized Gell-Mann basis
              - :meth:`partial` (not implemented)

    Since Basis is a subclass of NumPy's ``ndarray``, it inherits all of its
    attributes, e.g. ``shape``. The following attributes behave slightly
    differently to a ndarray, however

        - ``A == B`` is ``True`` if all elements evaluate almost equal, i.e.
          equivalent to ``np.allclose(A, B)``.
        - ``basis.T`` transposes the last two axes of ``basis``. For a full
          basis, this corresponds to transposing each element individually. For
          a basis element, it corresponds to normal transposition.

    Parameters
    ----------

    basis_array : array_like, shape (n, d, d)
        An array or list of square matrices that are elements of an operator
        basis spanning :math:`\mathbb{C}^{d\times d}`. *n* should be smaller
        than or equal to *d**2*.
    traceless : bool, optional (default: auto)
        Controls whether a traceless basis is forced. Here, traceless means
        that the first element of the basis is the identity and the remaining
        elements are matrices of trace zero. If an element of ``basis_array``
        is neither traceless nor the identity and ``traceless == True``, an
        exception will be raised. Defaults to ``True`` if basis_array is
        traceless and ``False`` if not.
    btype : str, optional (default: ``'custom'``)
        A string describing the basis type. For example, a basis created by the
        factory method :meth:`pauli` has *btype* 'pauli'.
    skip_check : bool, optional (default: ``False``)
        Skip the internal routine for checking ``basis_array``'s
        orthonormality and completeness. Use with caution.

    Attributes
    ----------
    Other than the attributes inherited from ``ndarray``, a ``Basis`` instance
    has the following attributes:

    btype : str
        Basis type.
    d : int
        Dimension of the space spanned by the basis.
    H : Basis
        Hermitian conjugate.
    isherm : bool
        If the basis is hermitian.
    isorthonorm : bool
        If the basis is orthonormal.
    istraceless : bool
        If the basis is traceless except for an identity element
    iscomplete : bool
        If the basis is complete, ie spans the full space.
    sparse : COO, shape (n, d, d)
        Representation in the COO format supplied by the ``sparse`` package.
    four_element_traces : COO, shape (n, n, n, n)
        Traces over all possible combinations of four elements of self. This is
        required for the calculation of the error transfer matrix and thus
        cached in the Basis instance.

    Most of the attributes above are properties which are lazily evaluated and
    cached.

    Methods
    -------
    Other than the methods inherited from ``ndarray``, a ``Basis`` instance has
    the following methods:

    :meth:`normalize`
        Normalizes the basis in-place (used internally when creating a basis
        from elements)
    :meth:`tidyup`
        Cleans up floating point errors in-place to make zeros actual zeros.
        ``eps_scale`` is an optional argument multiplied to the data type's
        ``eps`` to get the absolute tolerance.

    """

    def __new__(cls, basis_array: Sequence, traceless: Optional[bool] = None,
                btype: Optional[str] = None, skip_check: bool = False
                ) -> 'Basis':
        """Constructor."""
        if not skip_check:
            if not hasattr(basis_array, '__getitem__'):
                raise TypeError('Invalid data type. Must be array_like')

            if isinstance(basis_array, cls):
                basis = basis_array
            else:
                try:
                    # Allow single 2d element
                    if len(basis_array.shape) == 2:
                        basis_array = [basis_array]
                except AttributeError:
                    pass

                basis = np.empty((len(basis_array), *basis_array[0].shape),
                                 dtype=complex)
                for i, elem in enumerate(basis_array):
                    if isinstance(elem, ndarray):
                        basis[i] = elem
                    elif isinstance(elem, Qobj):
                        basis[i] = elem.full()
                    elif isinstance(elem, COO):
                        basis[i] = elem.todense()
                    else:
                        raise TypeError('At least one element invalid type!')

            d = basis.shape[-1]

            if len(basis) > d**2:
                raise ValueError('Given overcomplete set of basis matrices. '
                                 'Not linearly independent.')

            basis = _full_from_partial(basis, traceless)
        else:
            basis = np.asanyarray(basis_array)
            d = basis.shape[-1]

        basis = basis.view(cls)
        basis.btype = btype if btype is not None else 'Custom'
        basis.d = d
        return basis

    def __array_finalize__(self, basis: 'Basis') -> None:
        """Required for subclassing ndarray."""
        if basis is None:
            return

        self.btype = getattr(basis, 'btype', 'Custom')
        self.d = getattr(basis, 'd', basis.shape[-1])
        self._sparse = None     # sparse representation of self
        self._four_element_traces = None
        self._isherm = None
        self._isorthonorm = None
        self._istraceless = None
        self._iscomplete = None
        self._eps = np.finfo(complex).eps
        self._atol = self._eps*self.d**3
        self._rtol = 0

    def __eq__(self, other: object) -> bool:
        """Compare for equality."""
        try:
            if self.shape != other.shape:
                return False
        except AttributeError:
            # Not ndarray
            return np.equal(self, other)

        return np.allclose(self.view(ndarray), other.view(ndarray),
                           atol=self._atol, rtol=self._rtol)

    def __contains__(self, item: ndarray) -> bool:
        """Implement 'in' operator."""
        return any(np.all(np.isclose(
            item.view(ndarray), self.view(ndarray), rtol=self._rtol,
            atol=self._atol), axis=(1, 2)))

    def __array_wrap__(self, out_arr, context=None):
        """
        Fixes problem that ufuncs return 0-d arrays instead of scalars.

        https://github.com/numpy/numpy/issues/5819#issue-72454838
        """
        if out_arr.ndim:
            return ndarray.__array_wrap__(self, out_arr, context)

    def _print_checks(self) -> None:
        """Print checks for debug purposes."""
        checks = ('isherm', 'istraceless', 'iscomplete', 'isorthonorm')
        for check in checks:
            print(check, ':\t', getattr(self, check))

    @property
    def isherm(self) -> bool:
        """Returns True if all basis elements are hermitian."""
        if self._isherm is None:
            self._isherm = (self.H == self)

        return self._isherm

    @property
    def isorthonorm(self) -> bool:
        """Returns True if basis is orthonormal."""
        if self._isorthonorm is None:
            # All the basis is orthonormal iff the matrix consisting of all
            # d**2 elements written as d**2-dimensional column vectors is
            # unitary.
            if self.ndim == 2:
                # Only one basis element
                self._isorthonorm = True
            else:
                # Size of the result after multiplication
                dim = self.shape[0]
                U = self.reshape((dim, -1))
                actual = U.conj() @ U.T
                target = np.identity(dim)
                atol = self._eps*(self.d**2)**3
                self._isorthonorm = np.allclose(actual.view(ndarray), target,
                                                atol=atol, rtol=self._rtol)

        return self._isorthonorm

    @property
    def istraceless(self) -> bool:
        """
        Returns True if basis is traceless except for possibly the identity.
        """
        if self._istraceless is None:
            trace = np.einsum('...jj', self)
            trace = remove_float_errors(trace, self.d)
            nonzero = trace.nonzero()
            if nonzero[0].size == 0:
                self._istraceless = True
            elif nonzero[0].size == 1:
                # Single element has nonzero trace, check if (proportional to)
                # identity
                elem = self[nonzero][0].view(ndarray)
                offdiag_nonzero = elem[~np.eye(self.d, dtype=bool)].nonzero()
                diag_equal = np.diag(elem) == elem[0, 0]
                if diag_equal.all() and not offdiag_nonzero[0].any():
                    # Element is (proportional to) the identity, this we define
                    # as 'traceless' since a complete basis cannot have only
                    # traceless elems.
                    self._istraceless = True
                else:
                    # Element not the identity, therefore not traceless
                    self._istraceless = False
            else:
                self._istraceless = False

        return self._istraceless

    @property
    def iscomplete(self) -> bool:
        """Returns True if basis is complete."""
        if self._iscomplete is None:
            A = self.reshape(self.shape[0], -1)
            rank = np.linalg.matrix_rank(A)
            self._iscomplete = rank == self.d**2

        return self._iscomplete

    @property
    def H(self) -> 'Basis':
        """Return the basis hermitian conjugated element-wise."""
        return self.T.conj()

    @property
    def T(self) -> 'Basis':
        """Return the basis transposed element-wise."""
        if self.ndim == 3:
            return self.transpose(0, 2, 1)

        if self.ndim == 2:
            return self.transpose(1, 0)

        return self

    @property
    def sparse(self) -> COO:
        """Return the basis as a sparse COO array"""
        if self._sparse is None:
            self._sparse = COO.from_numpy(self)

        return self._sparse

    @property
    def four_element_traces(self) -> COO:
        r"""
        Return all traces of the form :math:`\mathrm{tr}(C_i C_j C_k C_l)` as
        a sparse COO array for :math:`i,j,k,l > 0` (i.e. excluding the
        identity).
        """
        if self._four_element_traces is None:
            # Most of the traces are zero, therefore store the result in a
            # sparse array. For GGM bases, which are inherently sparse, it
            # makes sense for any dimension to also calculate with sparse
            # arrays. For Pauli bases, which are very dense, this is not so
            # efficient but unavoidable for d > 12.
            path = [(0, 1), (0, 1), (0, 1)]
            if self.btype == 'Pauli' and self.d <= 12:
                # For d == 12, the result is ~270 MB.
                self._four_element_traces = COO.from_numpy(oe.contract(
                    'iab,jbc,kcd,lda->ijkl', *(self,)*4, optimize=path
                ))
            else:
                self._four_element_traces = oe.contract(
                    'iab,jbc,kcd,lda->ijkl', *(self.sparse,)*4,
                    backend='sparse', optimize=path
                )

        return self._four_element_traces

    @four_element_traces.setter
    def four_element_traces(self, traces):
        self._four_element_traces = traces

    def normalize(self) -> None:
        """Normalize the basis in-place"""
        if self.ndim == 2:
            self /= norm(self)
        elif self.ndim == 3:
            np.einsum('ijk,i->ijk', self, 1/norm(self, axis=(1, 2)),
                      out=self)

    def tidyup(self, eps_scale: Optional[float] = None) -> None:
        """Wraps util.remove_float_errors."""
        if eps_scale is None:
            atol = self._atol
        else:
            atol = self._eps*eps_scale

        self.real[np.abs(self.real) <= atol] = 0
        self.imag[np.abs(self.imag) <= atol] = 0

    @classmethod
    def partial(cls, basis_array, subspace_inds):
        raise NotImplementedError

    @classmethod
    def pauli(cls, n: int) -> 'Basis':
        r"""
        Returns a Pauli basis for :math:`n` qubits, i.e. the basis spans the
        space :math:`\mathbb{C}^{d\times d}` with :math:`d = 2^n`:

        .. math::
            \mathcal{P} = \{I, X, Y, Z\}^{\otimes n}.

        The elements :math:`\sigma_i` are normalized with respect to the
        Hilbert-Schmidt inner product,

        .. math::

            \langle\sigma_i,\sigma_j\rangle
                &= \mathrm{Tr}\,\sigma_i^\dagger\sigma_j \\
                &= \delta_{ij}.

        Parameters
        ----------
        n : int
            The number of qubits.

        Returns
        -------
        basis : Basis
            The Basis object representing the Pauli basis.
        """
        normalization = np.sqrt(2**n)
        combinations = np.indices((4,)*n).reshape(n, 4**n)
        sigma = tensor(*np.array(P_np)[combinations], rank=2)/normalization
        return cls(sigma, btype='Pauli', skip_check=True)

    @classmethod
    def ggm(cls, d: int) -> 'Basis':
        r"""
        Returns a generalized Gell-Mann basis in :math:`d` dimensions [Bert08]_
        where the elements :math:`\Lambda_i` are normalized with respect to the
        Hilbert-Schmidt inner product,

        .. math::

            \langle\Lambda_i,\Lambda_j\rangle
                &= \mathrm{Tr}\,\Lambda_i^\dagger\Lambda_j \\
                &= \delta_{ij}.

        Parameters
        ----------
        d : int
            The dimensionality of the space spanned by the basis

        Returns
        -------
        basis : Basis
            The Basis object representing the GGM.

        References
        ----------
        .. [Bert08]
            Bertlmann, R. A., & Krammer, P. (2008). Bloch vectors for qudits.
            Journal of Physics A: Mathematical and Theoretical, 41(23).
            https://doi.org/10.1088/1751-8113/41/23/235303

        """
        n_sym = int(d*(d - 1)/2)
        sym_rng = np.arange(1, n_sym + 1)
        diag_rng = np.arange(1, d)

        # Indices for offdiagonal elements
        j = np.repeat(np.arange(d - 1), np.arange(d - 1, 0, -1))
        k = np.arange(1, n_sym+1) - (j*(2*d - j - 3)/2).astype(int)
        j_offdiag = tuple(j)
        k_offdiag = tuple(k)
        # Indices for diagonal elements
        j_diag = tuple(i for l in range(d) for i in range(l))
        l_diag = tuple(i for i in range(1, d))

        inv_sqrt2 = 1/np.sqrt(2)
        Lambda = np.zeros((d**2, d, d), dtype=complex)
        Lambda[0] = np.eye(d)/np.sqrt(d)
        # First n matrices are symmetric
        Lambda[sym_rng, j_offdiag, k_offdiag] = inv_sqrt2
        Lambda[sym_rng, k_offdiag, j_offdiag] = inv_sqrt2
        # Second n matrices are antisymmetric
        Lambda[sym_rng+n_sym, j_offdiag, k_offdiag] = -1j*inv_sqrt2
        Lambda[sym_rng+n_sym, k_offdiag, j_offdiag] = 1j*inv_sqrt2
        # Remaining matrices have entries on the diagonal only
        Lambda[np.repeat(diag_rng, diag_rng)+2*n_sym, j_diag, j_diag] = 1
        Lambda[diag_rng + 2*n_sym, l_diag, l_diag] = -diag_rng
        # Normalize
        Lambda[2*n_sym + 1:, range(d), range(d)] /= \
            np.tile(np.sqrt(diag_rng*(diag_rng + 1))[:, None], (1, d))

        return cls(Lambda, btype='GGM', skip_check=True)


def _full_from_partial(elems: Sequence, traceless: Union[None, bool]) -> Basis:
    """
    Internal function to parse the basis elements *elems*. By default, checks
    are performed for orthogonality and linear independence. If either fails an
    exception is raised. Returns a full hermitian and orthonormal basis.
    """
    elems = np.asanyarray(elems)
    if not isinstance(elems, Basis):
        # Convert elems to basis to have access to its handy attributes
        elems = normalize(elems.view(Basis))

    if not elems.isherm:
        warn("(Some) elems not hermitian! The resulting basis also won't be.")

    if not elems.isorthonorm:
        raise ValueError("The basis elements are not orthonormal!")

    if traceless is None:
        traceless = elems.istraceless
    else:
        if traceless and not elems.istraceless:
            raise ValueError("The basis elements are not traceless (up to " +
                             "an identity element) but a traceless basis " +
                             "was requested!")

    # Get a Generalized Gell-Mann basis to expand in (fulfills the desired
    # properties hermiticity and orthonormality, and therefore also linear
    # combinations, ie basis expansions, of it will). Split off the identity so
    # that for traceless bases we can put it in the front.
    if traceless:
        Id, ggm = np.split(Basis.ggm(elems.d), [1])
    else:
        ggm = Basis.ggm(elems.d)

    coeffs = expand(elems, ggm, tidyup=True)

    # Throw out coefficient vectors that are all zero (should only happen for
    # the identity)
    coeffs = coeffs[(coeffs != 0).any(axis=1)]
    if coeffs.size != 0:
        # Get d**2 - len(coeffs) vectors spanning the nullspace of coeffs.
        # Those together with coeffs span the whole space, and therefore also
        # the linear combinations of GGMs weighted with the coefficients will
        # span the whole matrix space
        coeffs = np.concatenate((coeffs, null_space(coeffs).T))
        # Our new basis is given by linear combinations of GGMs with coeffs
        basis = np.einsum('ij,jkl', coeffs, ggm)
    else:
        # Resulting array is of size zero, i.e. we can just return the GGMs
        basis = ggm

    # Add the identity again and normalize the new basis
    if traceless:
        basis = np.concatenate((Id, basis)).view(Basis)
    else:
        basis = basis.view(Basis)

    # Clean up
    basis.tidyup()

    return basis


def normalize(b: Sequence) -> Basis:
    r"""
    Return a copy of the basis *b* normalized with respect to the Frobenius
    norm [Gol85]_:

        :math:`||A||_F = \left[\sum_{i,j} |a_{i,j}|^2\right]^{1/2}`

    or equivalently, with respect to the Hilbert-Schmidt inner product as
    implemented by :func:`~filter_functions.util.dot_HS`.

    References
    ----------
    .. [Gol85]
        G. H. Golub and C. F. Van Loan, *Matrix Computations*,
        Baltimore, MD, Johns Hopkins University Press, 1985, pg. 15

    """
    b = np.asanyarray(b)
    if b.ndim == 2:
        return (b/norm(b)).view(Basis)
    if b.ndim == 3:
        return np.einsum('ijk,i->ijk', b, 1/norm(b, axis=(1, 2))).view(Basis)

    raise ValueError('Expected b.ndim to be either 2 or 3, not ' +
                     '{}.'.format(b.ndim))


def expand(M: Union[ndarray, Basis], basis: Union[ndarray, Basis],
           normalized: bool = True, tidyup: bool = False) -> ndarray:
    r"""
    Expand the array *M* in the basis given by *basis*.

    Parameters
    ----------
    M : array_like
        The square matrix (d, d) or array of square matrices (..., d, d) to be
        expanded in *basis*
    basis : array_like
        The basis of shape (m, d, d) in which to expand.
    normalized : bool {True}
        Wether the basis is normalized.
    tidyup : bool {False}
        Whether to set values below the floating point eps to zero.

    Returns
    -------
    coefficients : ndarray
        The coefficient array with shape (..., m) or (m,) if *M* was 2-d

    Notes
    -----
    For an orthogonal matrix basis :math:`\mathcal{C} = \big\{C_k\in
    \mathbb{C}^{d\times d}: \langle C_k,C_l\rangle_\mathrm{HS} \propto
    \delta_{kl}\big\}_{k=0}^{d^2-1}` with the Hilbert-Schmidt inner product as
    implemented by :func:`~filter_functions.util.dot_HS` and
    :math:`M\in\mathbb{C}^{d\times d}`, the expansion of
    :math:`M` in terms of :math:`\mathcal{C}` is given by

    .. math::
        M &= \sum_j c_j C_j, \\
        c_j &= \frac{\mathrm{tr}\big(M C_j\big)}{\mathrm{tr}\big(C_j^2\big)}.

    """
    coefficients = np.einsum('...ij,bji->...b', np.asarray(M), basis)

    if not normalized:
        coefficients /= np.einsum('bij,bji->b', basis, basis).real

    return remove_float_errors(coefficients) if tidyup else coefficients


def ggm_expand(M: Union[ndarray, Basis], traceless: bool = False) -> ndarray:
    r"""
    Expand the matrix *M* in a Generalized Gell-Mann basis [Bert08]_. This
    function makes use of the explicit construction prescription of the basis
    and thus makes do without computing the expansion coefficients as the
    overlap between the matrix and each basis element.

    Parameters
    ----------
    M : array_like
        The square matrix (d, d) or array of square matrices (..., d, d) to be
        expanded in a GGM basis.
    traceless : bool (default: False)
        Include the basis element proportional to the identity in the
        expansion. If it is known beforehand that M is traceless, the
        corresponding coefficient is zero and thus doesn't need to be computed.

    Returns
    -------
    coefficients : ndarray
        The coefficient array with shape (d**2,) or (..., d**2)

    References
    ----------
    .. [Bert08]
        Bertlmann, R. A., & Krammer, P. (2008). Bloch vectors for qudits.
        Journal of Physics A: Mathematical and Theoretical, 41(23).
        https://doi.org/10.1088/1751-8113/41/23/235303
    """
    if M.shape[-1] != M.shape[-2]:
        raise ValueError('M should be square in its last two axes')

    # Squeeze out an extra dimension to be shape agnostic
    square = M.ndim < 3
    if square:
        M = M[None, ...]

    d = M.shape[-1]

    n_sym = int(d*(d - 1)/2)
    sym_rng = np.arange(1, n_sym + 1)
    diag_rng = np.arange(1, d)

    # Map linear index to index tuple of upper triangle
    j = np.repeat(np.arange(d - 1), np.arange(d - 1, 0, -1))
    k = np.arange(1, n_sym+1) - (j*(2*d - j - 3)/2).astype(int)
    offdiag_idx = [tuple(j), tuple(k)]
    # Indices for upper triangular part of M
    triu_idx = tuple([...] + offdiag_idx)
    # Indices for lower triangular part of M
    tril_idx = tuple([...] + offdiag_idx[::-1])
    # Indices for diagonal elements of M up to first to last
    diag_idx = tuple([...] + [tuple(i for i in range(d - 1))]*2)
    # Indices of diagonal elements of M starting from first
    diag_idx_shifted = tuple([...] + [tuple(i for i in range(1, d))]*2)

    # Compute the coefficients
    coeffs = np.zeros((*M.shape[:-2], d**2), dtype=complex)
    if not traceless:
        # First element is proportional to the trace of M
        coeffs[..., 0] = np.einsum('...jj', M)/np.sqrt(d)
    # Elements proportional to the symmetric GGMs
    coeffs[..., sym_rng] = (M[triu_idx] + M[tril_idx])/np.sqrt(2)
    # Elements proportional to the antisymmetric GGMs
    coeffs[..., sym_rng + n_sym] = 1j*(M[triu_idx] - M[tril_idx])/np.sqrt(2)
    # Elements proportional to the diagonal GGMs
    coeffs[..., diag_rng + 2*n_sym] = (M[diag_idx].cumsum(axis=-1) -
                                       diag_rng*M[diag_idx_shifted])
    coeffs[..., diag_rng + 2*n_sym] /= np.sqrt(diag_rng*(diag_rng + 1))

    return coeffs.squeeze() if square else coeffs


def equivalent_pauli_basis_elements(idx: Union[Sequence[int], int],
                                    N: int) -> ndarray:
    """
    Get the indices of the equivalent (up to identities tensored to it)
    basis elements of Pauli bases of qubits at position idx in the total
    Pauli basis for N qubits.
    """
    idx = [idx] if isinstance(idx, int) else idx
    multi_index = np.ix_(*[range(4) if i in idx else [0]
                           for i in range(N)])
    elem_idx = np.ravel_multi_index(multi_index, [4]*N).ravel()
    return elem_idx


def remap_pauli_basis_elements(order: Sequence[int], N: int) -> ndarray:
    """
    For a N-qubit Pauli basis, transpose the order of the subsystems and return
    the indices that permute the old basis to the new.
    """
    # Index tuples for single qubit paulis that give the n-qubit paulis when
    # tensored together
    pauli_idx = np.indices((4,)*N).reshape(N, 4**N).T
    # Indices of the N-qubit basis reordered according to order
    linear_idx = [np.ravel_multi_index([idx_tup[i] for i in order], (4,)*N)
                  for idx_tup in pauli_idx]

    return np.array(linear_idx)



# -*- coding: utf-8 -*-
# =============================================================================
#     filter_functions
#     Copyright (C) 2019 Quantum Technology Group, RWTH Aachen University
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#     Contact email: tobias.hangleiter@rwth-aachen.de
# =============================================================================
"""
This module provides various plotting functions.

Functions
---------
:func:`plot_bloch_vector_evolution`
    Plot the evolution of the Bloch vector on a QuTiP-generated Bloch sphere
:func:`plot_filter_function`
    Plot the filter function of a given ``PulseSequence``
:func:`plot_infidelity_convergence`
    Helper function called by
    :func:`~filter_functions.pulse_sequence.infidelity` to plot the convergence
    of the infidelity
:func:`plot_pulse_correlation_filter_function`
    Plot the pulse correlation filter function of a given ``PulseSequence``
:func:`plot_pulse_train`
    Plot the pulse train of a given ``PulseSequence``
:func:`plot_error_transfer_matrix`
    Plot the error transfer matrix of a ``PulseSequence`` for a given spectrum
    as an image.

"""
from itertools import product
from typing import Optional, Sequence, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colors, lines
from mpl_toolkits.axes_grid1 import ImageGrid
from numpy import ndarray
from qutip import Bloch, Qobj, basis, expect

from . import numeric, util
from .types import (Axes, Coefficients, Colormap, Figure, FigureAxes,
                    FigureAxesLegend, FigureGrid, Grid, Operator, State)

__all__ = ['plot_bloch_vector_evolution', 'plot_error_transfer_matrix',
           'plot_filter_function', 'plot_infidelity_convergence',
           'plot_pulse_correlation_filter_function', 'plot_pulse_train']


def get_bloch_vector(states: Sequence[State]) -> ndarray:
    r"""
    Get the Bloch vector from quantum states.
    """
    if isinstance(states[0], Qobj):
        a = np.empty((3, len(states)))
        X, Y, Z = util.P_qt[1:]
        for i, state in enumerate(states):
            a[:, i] = [expect(X, state),
                       expect(Y, state),
                       expect(Z, state)]
    else:
        a = np.einsum('...ij,kil,...lm->k...', np.conj(states), util.P_np[1:],
                      states)
    return a


def init_bloch_sphere(**bloch_kwargs) -> Bloch:
    """A helper function to create a Bloch instance with a default viewing
    angle and axis labels."""
    bloch_kwargs.setdefault('view', [-150, 30])
    b = Bloch(**bloch_kwargs)
    b.xlabel = [r'$|+\rangle$', '']
    b.ylabel = [r'$|+_i\rangle$', '']
    return b


def get_states_from_prop(U: Sequence[Operator],
                         psi0: State = basis(2, 0),
                         prop: str = 'total') -> ndarray:
    r"""
    Get the the quantum state at time t from the propagator and the inital
    state:

    .. math::

        |\psi(t)\rangle = U(t, 0)|\psi(0)\rangle

    If *prop* is 'piecewise', then it is assumed that *U* is the propagator
    of a piecewise-constant control:

    .. math::
        |\psi(t)\rangle = \prod_{l=1}^n U(t_l, t_{l-1})|\psi(0)\rangle

    with :math:`t_0\equiv 0` and :math:`t_n\equiv t`.

    """
    psi0 = psi0.full() if isinstance(psi0, Qobj) else psi0
    d = max(psi0.shape)
    states = np.empty((len(U), d, 1), dtype=complex)
    if prop == 'total':
        for j in range(len(U)):
            states[j] = U[j] @ psi0
    elif prop == 'piecewise':
        states[0] = U[0] @ psi0
        for j in range(1, len(U)):
            states[j] = U[j] @ states[j-1]

    return states


def plot_bloch_vector_evolution(pulse: 'PulseSequence',
                                psi0: Optional[State] = None,
                                b: Optional[Bloch] = None,
                                n_samples: Optional[int] = None,
                                show: bool = True,
                                return_Bloch: bool = False,
                                **bloch_kwargs) -> Union[None, Bloch]:
    r"""
    Plot the evolution of the Bloch vector under the given pulse sequence.

    Parameters
    ----------
    pulse : PulseSequence
        The PulseSequence instance whose control Hamiltonian determines the
        time evolution of the Bloch vector.
    psi0 : Qobj or array_like, optional
        The initial state before the pulse is applied. Defaults to
        :math:`|0\rangle`.
    b : Bloch, optional
        If given, the QuTiP Bloch instance on which to plot the time evolution.
    n_samples : int, optional
        The number of time points to be sampled.
    show** : bool, optional
        Whether to show the sphere (by calling :code:`b.make_sphere()`).
    return_Bloch : bool, optional
        Whether to return the ``qutip.Bloch`` instance
    bloch_kwargs : dict, optional
        A dictionary with keyword arguments to be fed into the Bloch
        constructor (if *b* not given).

    Returns
    -------
    b : Bloch
        The Bloch instance

    Raises
    ------
    ValueError
        If the pulse is for more than one qubit

    See Also
    --------
    :class:`qutip.Bloch`
    """
    # Raise an exception if not a one-qubit pulse
    if not pulse.d == 2:
        raise ValueError('Plotting Bloch sphere evolution only implemented ' +
                         'for one-qubit case!')

    # Parse default arguments
    if b is None:
        b = init_bloch_sphere(**bloch_kwargs)

    if n_samples is None:
        # 5 time points during  the smallest time interval in pulse.t. Being
        # careful that doesn't blow up in our face for extremely narrow pulses,
        # max out at 5000.
        n_samples = min([5000, 5*int(pulse.t[-1]/np.diff(pulse.t).min())])

    if psi0 is None:
        psi0 = basis(2)

    times = np.linspace(pulse.t[0], pulse.t[-1], n_samples)
    n_cops = len(pulse.c_opers)
    coeffs = np.zeros((n_cops, len(times)))
    for i in range(n_cops):
        for j, coeff in enumerate(pulse.c_coeffs[i]):
            if coeff != 0:
                coeffs[i] += coeff*(pulse.t[j] <= times) *\
                    (times <= pulse.t[j+1])

    propagators = pulse.propagator_at_arb_t(times)
    points = get_bloch_vector(get_states_from_prop(propagators, psi0))
    b.add_points(points, meth='l')
    if show:
        b.make_sphere()

    if return_Bloch:
        return b

    return


def plot_pulse_train(pulse: 'PulseSequence',
                     c_oper_identifiers: Optional[Sequence[int]] = None,
                     fig: Optional[Figure] = None,
                     axes: Optional[Axes] = None,
                     plot_kw: Optional[dict] = {},
                     subplot_kw: Optional[dict] = None,
                     gridspec_kw: Optional[dict] = None,
                     **figure_kw) -> FigureAxesLegend:
    """
    Plot the pulsetrain of the ``PulseSequence`` *pulse*.

    Parameters
    ----------
    pulse : PulseSequence
        The pulse sequence whose pulse train to plot.
    n_oper_identifiers : array_like, optional
        The identifiers of the control operators for which the pulse train
        should be plotted. All identifiers can be accessed via
        ``pulse.c_oper_identifiers``. Defaults to all.
    fig : matplotlib figure, optional
        A matplotlib figure instance to plot in
    axes : matplotlib axes, optional
        A matplotlib axes instance to use for plotting.
    plot_kw : dict, optional
        Dictionary with keyword arguments passed to the plot function
    subplot_kw : dict, optional
        Dictionary with keyword arguments passed to the subplots constructor
    gridspec_kw : dict, optional
        Dictionary with keyword arguments passed to the gridspec constructor
    figure_kw : optional
        Keyword argument dictionaries that are fed into the
        :func:`matplotlib.pyplot.subplots` function if no *fig* instance is
        specified.

    Returns
    -------
    fig : matplotlib figure
        The matplotlib figure instance used for plotting.
    axes : matplotlib axes
        The matplotlib axes instance used for plotting.
    legend : matplotlib legend
        The matplotlib legend instance in the plot.

    Raises
    ------
    ValueError
        If an invalid number of c_oper_labels were given
    """
    c_oper_inds = util.get_indices_from_identifiers(pulse, c_oper_identifiers,
                                                    'control')
    c_oper_identifiers = pulse.c_oper_identifiers[c_oper_inds]

    if fig is None and axes is None:
        fig, axes = plt.subplots(subplot_kw=subplot_kw,
                                 gridspec_kw=gridspec_kw,
                                 **figure_kw)
    elif axes is None and fig is not None:
        subplot_kw = {} if subplot_kw is None else subplot_kw
        axes = fig.add_subplot(111, **subplot_kw)
    elif fig is None and axes is not None:
        fig = axes.figure

    handles = []
    for i, c_coeffs in enumerate(pulse.c_coeffs[tuple(c_oper_inds), ...]):
        coeffs = np.insert(c_coeffs, 0, c_coeffs[0])
        handles += axes.step(pulse.t, coeffs, label=c_oper_identifiers[i],
                             **plot_kw)

    axes.set_xlim(pulse.t[0], pulse.t[-1])
    axes.set_xlabel(r'$t$ / a.u.')
    axes.set_ylabel(r'Control parameter / a.u.')
    axes.grid(True)
    legend = axes.legend(framealpha=1)

    return fig, axes, legend


def plot_filter_function(pulse: 'PulseSequence',
                         omega: Optional[Coefficients] = None,
                         n_oper_identifiers: Optional[Sequence[int]] = None,
                         fig: Optional[Figure] = None,
                         axes: Optional[Axes] = None,
                         xscale: str = 'log',
                         yscale: str = 'linear',
                         omega_in_units_of_tau: bool = True,
                         plot_kw: dict = {},
                         subplot_kw: Optional[dict] = None,
                         gridspec_kw: Optional[dict] = None,
                         **figure_kw) -> FigureAxesLegend:
    r"""
    Plot the filter function(s) of the given PulseSequence for positive
    frequencies. As of now only the diagonal elements of
    :math:`F_{\alpha\beta}` are implemented, i.e. the filter functions
    corresponding to uncorrelated noise sources.

    Parameters
    ----------
    pulse : PulseSequence
        The pulse sequence whose filter function to plot.
    omega : array_like, optional
        The frequencies at which to evaluate the filter function. If not given,
        the pulse sequence's omega attribute is used (if set) or sensible
        values are chosen automatically (if ``None``)
    n_oper_identifiers : array_like, optional
        The identifiers of the noise operators for which the filter function
        should be plotted. All identifiers can be accessed via
        ``pulse.n_oper_identifiers``. Defaults to all.
    fig : matplotlib figure, optional
        A matplotlib figure instance to plot in
    axes : matplotlib axes, optional
        A matplotlib axes instance to use for plotting.
    xscale : str, optional
        x-axis scaling. One of ('linear', 'log').
    yscale : str, optional
        y-axis scaling. One of ('linear', 'log').
    omega_in_units_of_tau : bool, optional
        Plot :math:`\omega\tau` or just :math:`\omega` on x-axis.
    plot_kw : dict, optional
        Dictionary with keyword arguments passed to the plot function
    subplot_kw : dict, optional
        Dictionary with keyword arguments passed to the subplots constructor
    gridspec_kw : dict, optional
        Dictionary with keyword arguments passed to the gridspec constructor
    figure_kw : optional
        Keyword argument dictionaries that are fed into the
        :func:`matplotlib.pyplot.subplots` function if no *fig* instance is
        specified.

    Returns
    -------
    fig : matplotlib figure
        The matplotlib figure instance used for plotting.
    axes : matplotlib axes
        The matplotlib axes instance used for plotting.
    legend : matplotlib legend
        The matplotlib legend instance in the plot.

    Raises
    ------
    ValueError
        If an invalid number of n_oper_labels were given
    """
    if omega is None:
        if pulse.omega is None:
            omega = util.get_sample_frequencies(pulse, spacing=xscale)
        else:
            omega = pulse.omega

    n_oper_inds = util.get_indices_from_identifiers(pulse, n_oper_identifiers,
                                                    'noise')
    n_oper_identifiers = pulse.n_oper_identifiers[n_oper_inds]

    if fig is None and axes is None:
        fig, axes = plt.subplots(subplot_kw=subplot_kw,
                                 gridspec_kw=gridspec_kw,
                                 **figure_kw)
    elif axes is None and fig is not None:
        subplot_kw = {} if subplot_kw is None else subplot_kw
        axes = fig.add_subplot(111, **subplot_kw)
    elif fig is None and axes is not None:
        fig = axes.figure

    if omega_in_units_of_tau:
        tau = np.ptp(pulse.t)
        z = omega*tau
        xlabel = r'$\omega\tau$'
    else:
        z = omega
        xlabel = r'$\omega$'

    diag_idx = np.arange(len(pulse.n_opers))
    F = pulse.get_filter_function(omega)[diag_idx, diag_idx].real

    handles = []
    for i, ind in enumerate(n_oper_inds):
        handles += axes.plot(z, F[ind], label=n_oper_identifiers[i], **plot_kw)

    # Set the axis scales
    axes.set_xscale(xscale)
    axes.set_yscale(yscale)
    if xscale == 'log':
        z_min_idx = (z > 0).nonzero()[0][0]
    else:
        z_min_idx = (z >= 0).nonzero()[0][0]

    if yscale != 'log':
        axes.set_ylim(bottom=0)

    axes.set_xlim(z[z_min_idx], max(z))
    axes.set_xlabel(xlabel)
    axes.set_ylabel(r'$F(\omega)$')
    axes.grid(True)
    legend = axes.legend()

    return fig, axes, legend


def plot_pulse_correlation_filter_function(
        pulse: 'PulseSequence',
        n_oper_identifiers: Optional[Sequence[int]] = None,
        fig: Optional[Figure] = None,
        xscale: str = 'log',
        yscale: str = 'linear',
        omega_in_units_of_tau: bool = True,
        plot_kw: dict = {},
        subplot_kw: Optional[dict] = None,
        gridspec_kw: Optional[dict] = None,
        **figure_kw) -> FigureAxesLegend:
    r"""
    Plot the pulse correlation filter functions of the given PulseSequence if
    they were computed during concatenation for positive frequencies.

    Returns a figure with *n* by *n* subplots where *n* is the number of pulses
    that were concatenated. As of now only the diagonal elements of
    :math:`F_{\alpha\beta}` are implemented, i.e. the filter functions
    corresponding to uncorrelated noise sources.

    Parameters
    ----------
    pulse : PulseSequence
        The pulse sequence whose filter function to plot.
    n_oper_identifiers : array_like, optional
        The identifiers of the noise operators for which the filter function
        should be plotted. All identifiers can be accessed via
        ``pulse.n_oper_identifiers``. Defaults to all.
    fig : matplotlib figure, optional
        A matplotlib figure instance to plot in
    xscale : str, optional
        x-axis scaling. One of ('linear', 'log').
    yscale : str, optional
        y-axis scaling. One of ('linear', 'log').
    omega_in_units_of_tau : bool, optional
        Plot :math:`\omega\tau` or just :math:`\omega` on x-axis.
    plot_kw : dict, optional
        Dictionary with keyword arguments passed to the plot function
    subplot_kw : dict, optional
        Dictionary with keyword arguments passed to the subplots constructor
    gridspec_kw : dict, optional
        Dictionary with keyword arguments passed to the gridspec constructor
    figure_kw : optional
        Keyword argument dictionaries that are fed into the
        :func:`matplotlib.pyplot.subplots` function if no *fig* instance is
        specified.

    Returns
    -------
    fig : matplotlib figure
        The matplotlib figure instance used for plotting.
    axes : matplotlib axes
        The matplotlib axes instances used for plotting.
    legend : matplotlib legend
        The matplotlib legend instance in the plot.

    Raises
    ------
    CalculationError
        If the pulse correlation filter function was not computed during
        concatenation.
    """
    n_oper_inds = util.get_indices_from_identifiers(pulse, n_oper_identifiers,
                                                    'noise')
    n_oper_identifiers = pulse.n_oper_identifiers[n_oper_inds]
    diag_idx = np.arange(len(pulse.n_opers))
    F_pc = pulse.get_pulse_correlation_filter_function()
    F_pc = F_pc[:, :, diag_idx, diag_idx]
    n = F_pc.shape[0]

    if fig is None:
        fig, axes = plt.subplots(n, n, sharex=True, subplot_kw=subplot_kw,
                                 gridspec_kw=gridspec_kw,
                                 **figure_kw)

    else:
        subplot_kw = {} if subplot_kw is None else subplot_kw
        axes = np.empty((n, n), dtype='O')
        axes[0, 0] = fig.add_subplot(n, n, 1, **subplot_kw)
        for row in range(n):
            for col in range(n):
                if not (row == 0 and col == 0):
                    index = np.ravel_multi_index([row, col], dims=(n, n)) + 1
                    axes[row, col] = fig.add_subplot(n, n, index,
                                                     sharex=axes[0, 0],
                                                     **subplot_kw)

    omega = pulse.omega
    if omega_in_units_of_tau:
        tau = np.ptp(pulse.t)
        z = omega*tau
        xlabel = r'$\omega\tau$'
    else:
        z = omega
        xlabel = r'$\omega$'

    transparent_line = lines.Line2D([], [], alpha=0)
    solid_line = lines.Line2D([], [], color='gray', linestyle='-')
    dashed_line = lines.Line2D([], [], color='gray', linestyle='--')
    for i in range(n):
        for j in range(n):
            handles = []
            for k, ind in enumerate(n_oper_inds):
                handles += axes[i, j].plot(z, F_pc[i, j, ind].real,
                                           label=n_oper_identifiers[k],
                                           **plot_kw)
                if i != j:
                    axes[i, j].plot(z, F_pc[i, j, ind].imag, linestyle='--',
                                    color=handles[-1].get_color(), **plot_kw)

            # Set the axis scales
            axes[i, j].set_yscale(yscale)
            axes[i, j].set_title(r'$F^{{({}{})}}(\omega)$'.format(i, j))
            if i != n-1:
                # Hide the ticklabels on all but the lowest row
                axes[i, j].axes.xaxis.set_ticklabels([])
            else:
                axes[i, j].set_xlabel(xlabel)

            if i == 0 and j == n-1:
                handles += [transparent_line, solid_line, dashed_line]
                labels = n_oper_identifiers.tolist() + ['', r'$Re$', r'$Im$']
                legend = axes[i, j].legend(handles=handles, labels=labels,
                                           bbox_to_anchor=(1.05, 1), loc=2,
                                           borderaxespad=0., frameon=False)

    axes[i, j].set_xscale(xscale)
    if xscale == 'log':
        z_min_idx = (z > 0).nonzero()[0][0]
    else:
        z_min_idx = (z >= 0).nonzero()[0][0]

    axes[i, j].set_xlim(z[z_min_idx], max(z))
    fig.tight_layout(h_pad=0.1)

    return fig, axes, legend


def plot_infidelity_convergence(n_samples: Sequence[int],
                                infids: Sequence[float]) -> FigureAxes:
    """
    Plot the convergence of the infidelity integral. The function arguments are
    those returned by :func:`~filter_functions.numeric.infidelity` with the
    *test_convergence* flag set to ``True``.

    Parameters
    ----------
    n_samples : array_like
        Array with the number of samples at which the integral was evaluated
    infids : array_like, shape (n_samples, [n_oper_inds, optional])
        Array with the calculated infidelities for each noise operator on the
        second axis or the second axis already traced out.

    Returns
    -------
    fig : matplotlib figure
        The matplotlib figure instance used for plotting.
    axes : matplotlib axes
        The matplotlib axes instances used for plotting.

    """
    fig, ax = plt.subplots(2, 1, sharex=True)
    ax[1].set_xlabel(r'$n_\omega$')
    ax[0].set_ylabel(r'$\mathcal{I}$')
    ax[1].set_ylabel(r'$|\Delta\mathcal{I}|/\mathcal{I}$ (%)')
    ax[0].set_xlim(min(n_samples), max(n_samples))
    ax[0].grid()
    ax[1].grid()

    ax[0].plot(n_samples, infids, 'o-')
    ax[1].semilogy(n_samples[1:],
                   np.abs(np.diff(infids, axis=-1))/infids[1:]*100, 'o-')

    return fig, ax


def plot_error_transfer_matrix(
        pulse: Optional['PulseSequence'] = None,
        S: Optional[ndarray] = None,
        omega: Optional[Coefficients] = None,
        U: Optional[ndarray] = None,
        n_oper_identifiers: Optional[Sequence[int]] = None,
        basis_labels: Optional[Sequence[str]] = None,
        colorscale: Optional[str] = 'linear',
        linthresh: Optional[float] = None,
        cbar_label: Optional[str] = 'Error transfer matrix',
        basis_labelsize: Optional[int] = None,
        fig: Optional[Figure] = None,
        grid: Optional[Grid] = None,
        cmap: Optional[Colormap] = None,
        grid_kw: Optional[dict] = None,
        imshow_kw: Optional[dict] = None,
        **figure_kw) -> FigureGrid:
    """
    Plot the error transfer matrix for a given noise spectrum as an image.

    The function may be called with either a ``PulseSequence``, a spectrum, and
    a list of frequencies in which case the error transfer matrix is calculated
    for those parameters, or with a precomputed error transfer matrix.

    As of now, only auto-correlated spectra are implemented.

    Parameters
    ----------
    pulse : 'PulseSequence'
        The pulse sequence.
    S : ndarray
        The two-sided noise spectrum.
    omega : array_like
        The frequencies for which to evaluate the error transfer matrix. Note
        that they should be symmetric around zero, that is, include negative
        frequencies.
    U : ndarray, shape
        A precomputed error transfer matrix. If given, *pulse*, *S*, *omega*
        are not required.
    n_oper_identifiers : array_like, optional
        The identifiers of the noise operators for which the error transfer
        matrix should be plotted. All identifiers can be accessed via
        ``pulse.n_oper_identifiers``. Defaults to all.
    basis_labels : array_like (str), optional
        Labels for the elements of the error transfer matrix (the basis
        elements).
    colorscale : str, optional
        The scale of the color code ('linear' or 'log' (default))
    linthresh: float, optional
        The threshold below which the colorscale will be linear (only for
        'log') colorscale
    cbar_label : str, optional
        The label for the colorbar. Default: 'Error transfer matrix'.
    basis_labelsize : int, optional
        The size in points for the basis labels.
    fig : matplotlib figure, optional
        A matplotlib figure instance to plot in
    grid : matplotlib ImageGrid, optional
        An ImageGrid instance to use for plotting.
    cmap : matplotlib colormap, optional
        The colormap for the matrix plot.
    subplot_kw : dict, optional
        Dictionary with keyword arguments passed to the ImageGrid constructor.
    imshow_kw : dict, optional
        Dictionary with keyword arguments passed to imshow.
    figure_kw : optional
        Keyword argument dictionaries that are fed into the
        :func:`matplotlib.pyplot.figure` function if no *fig* instance is
        specified.

    Returns
    -------
    fig : matplotlib figure
        The matplotlib figure instance used for plotting.
    grid : matplotlib ImageGrid
        The ImageGrid instance used for plotting.
    """
    if U is not None:
        if U.ndim == 2:
            U = np.array([U])

        n_oper_inds = np.arange(len(U))
        if n_oper_identifiers is None:
            if pulse is not None and len(pulse.n_oper_identifiers) == len(U):
                n_oper_identifiers = pulse.n_oper_identifiers
            else:
                n_oper_identifiers = ['$B_{}$'.format(i)
                                      for i in range(len(n_oper_inds))]

    else:
        if pulse is None or S is None or omega is None:
            raise ValueError('Require either precomputed error transfer ' +
                             'matrix or pulse, S, and omega as arguments.')

        n_oper_inds = util.get_indices_from_identifiers(pulse,
                                                        n_oper_identifiers,
                                                        'noise')
        n_oper_identifiers = pulse.n_oper_identifiers[n_oper_inds]
        # Get the error transfer matrix
        U = numeric.error_transfer_matrix(pulse, S, omega, n_oper_identifiers)
        if U.ndim == 4:
            # Only autocorrelated noise supported
            U = U[range(len(n_oper_inds)), range(len(n_oper_inds))]

    # Only autocorrelated noise implemented for now, ie U is real
    U = U.real

    if basis_labels is None:
        btype = pulse.basis.btype if pulse is not None else ''
        if btype == 'Pauli':
            n_qubits = int(np.log(U.shape[-1])/np.log(4))
            basis_labels = [''.join(tup) for tup in
                            product(['I', 'X', 'Y', 'Z'], repeat=n_qubits)]
        else:
            basis_labels = ['$C_{{{}}}$'.format(i)
                            for i in range(U.shape[-1])]
    else:
        if len(basis_labels) != U.shape[-1]:
            raise ValueError('Invalid number of basis_labels given')

    if grid is None:
        aspect_ratio = 2/3
        n_rows = int(np.round(np.sqrt(aspect_ratio*len(n_oper_inds))))
        n_cols = int(np.ceil(len(n_oper_inds)/n_rows))
        grid_kw = {} if grid_kw is None else grid_kw
        grid_kw.setdefault('rect', 111)
        grid_kw.setdefault('nrows_ncols', (n_rows, n_cols))
        grid_kw.setdefault('axes_pad', 0.3)
        grid_kw.setdefault('label_mode', 'L')
        grid_kw.setdefault('add_all', True)
        grid_kw.setdefault('share_all', True)
        grid_kw.setdefault('direction', 'row')
        grid_kw.setdefault('cbar_mode', 'single')
        grid_kw.setdefault('cbar_pad', 0.3)
        if fig is None:
            figsize = figure_kw.pop('figsize', (8*n_cols, 6*n_rows))
            fig = plt.figure(figsize=figsize, **figure_kw)

        grid = ImageGrid(fig, **grid_kw)
    else:
        if len(grid) != len(n_oper_inds):
            raise ValueError('Size of supplied ImageGrid instance does not ' +
                             'match the number of n_oper_identifiers given!')

        fig = grid[0].get_figure()

    # Parse default arguments
    if cmap is not None:
        plt.get_cmap(cmap)
    else:
        cmap = plt.get_cmap('RdBu')

    Umax = U.max()
    Umin = -Umax
    if colorscale == 'log':
        linthresh = np.abs(U).mean()/10 if linthresh is None else linthresh
        norm = colors.SymLogNorm(linthresh=linthresh, vmin=Umin, vmax=Umax)
    elif colorscale == 'linear':
        norm = colors.Normalize(vmin=Umin, vmax=Umax)

    imshow_kw = {} if imshow_kw is None else imshow_kw
    imshow_kw.setdefault('origin', 'upper')
    imshow_kw.setdefault('interpolation', 'nearest')
    imshow_kw.setdefault('cmap', cmap)
    imshow_kw.setdefault('norm', norm)

    basis_labelsize = 8 if basis_labelsize is None else basis_labelsize

    # Draw the images
    for i, n_oper_identifier in enumerate(n_oper_identifiers):
        ax = grid[i]
        im = ax.imshow(U[i], **imshow_kw)
        ax.set_title(n_oper_identifier)
        ax.set_xticks(np.arange(U.shape[-1]))
        ax.set_yticks(np.arange(U.shape[-1]))
        ax.set_xticklabels(basis_labels, rotation='vertical',
                           fontsize=basis_labelsize)
        ax.set_yticklabels(basis_labels, fontsize=basis_labelsize)
        ax.spines['left'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.spines['bottom'].set_visible(False)

    # Set up the colorbar
    cbar = fig.colorbar(im, cax=grid.cbar_axes[0])
    cbar.set_label(cbar_label)
    if colorscale == 'log':
        labels = cbar.ax.get_yticklabels()
        labels[len(labels) // 2] = ''
        labels = cbar.ax.set_yticklabels(labels)

    return fig, grid



# -*- coding: utf-8 -*-
# =============================================================================
#     filter_functions
#     Copyright (C) 2019 Quantum Technology Group, RWTH Aachen University
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#     Contact email: tobias.hangleiter@rwth-aachen.de
# =============================================================================
"""
This module provides various helper functions.

Functions
---------
:func:`abs2`
    Absolute value squared
:func:`get_indices_from_identifiers`
    The the indices of control or noise operators with given identifiers as
    they are saved in a ``PulseSequence``.
:func:`tensor`
    Fast, flexible tensor product of an arbitrary number of inputs using
    :func:`~numpy.einsum`
:func:`tensor_insert`
    For an array that is known to be a tensor product, insert arrays at a given
    position in the product chain
:func:`tensor_merge`
    For two arrays that are tensor products of known dimensions, merge them
    at arbitary positions in the product chain
:func:`tensor_transpose`
    For a tensor product, transpose the order of the constituents in the
    product chain
:func:`mdot`
    Multiple matrix product
:func:`remove_float_errors`
    Set entries whose absolute value is below a certain threshold to zero
:func:`oper_equiv`
    Determine if two vectors or operators are equal up to a global phase
:func:`dot_HS`
    Hilbert-Schmidt inner product
:func:`get_sample_frequencies`
    Get frequencies with typical infrared and ultraviolet cutoffs for a
    ``PulseSequence``
:func:`symmetrize_spectrum`
    Symmetrize a one-sided power spectrum as well as the frequencies associated
    with it to get a two-sided spectrum.
:func:`progressbar`
    A progress bar for loops. Uses tqdm if available and a simple custom one if
    not.
:func:`hash_array_along_axis`
    Return a list of hashes along a given axis
:func:`all_array_equal`
    Check if all arrays in an iterable are equal

Exceptions
----------
:class:`CalculationError`
    Exception raised if trying to fetch the pulse correlation function when it
    was not computed during concatenation

"""
import io
import json
import operator
import os
import re
import string
import sys
from functools import reduce
from itertools import zip_longest
from typing import Generator, Iterable, List, Optional, Sequence, Tuple, Union

import numpy as np
import qutip as qt
from numpy import linalg, ndarray

from .types import Operator, State

try:
    import jupyter_client
    import requests
    from jupyter_core.paths import jupyter_runtime_dir
    from notebook.utils import check_pid
    from requests.compat import urljoin

    def _list_running_servers(runtime_dir: str = None) -> Generator:
        """Iterate over the server info files of running notebook servers.

        Given a runtime directory, find nbserver-* files in the security
        directory, and yield dicts of their information, each one pertaining to
        a currently running notebook server instance.

        Copied from notebook.notebookapp.list_running_servers() (version 5.7.8)
        since the highest version compatible with Python 3.5 (version 5.6.0)
        has a bug.
        """
        if runtime_dir is None:
            runtime_dir = jupyter_runtime_dir()

        # The runtime dir might not exist
        if not os.path.isdir(runtime_dir):
            return

        for file_name in os.listdir(runtime_dir):
            if re.match('nbserver-(.+).json', file_name):
                with io.open(os.path.join(runtime_dir, file_name),
                             encoding='utf-8') as f:
                    info = json.load(f)

                # Simple check whether that process is really still running
                # Also remove leftover files from IPython 2.x without a pid
                # field
                if ('pid' in info) and check_pid(info['pid']):
                    yield info
                else:
                    # If the process has died, try to delete its info file
                    try:
                        os.unlink(os.path.join(runtime_dir, file_name))
                    except OSError:
                        pass  # TODO: This should warn or log or something

    def _get_notebook_name() -> str:
        """
        Return the full path of the jupyter notebook.

        See https://github.com/jupyter/notebook/issues/1000
        """
        try:
            connection_file = jupyter_client.find_connection_file()
        except OSError:
            return ''

        kernel_id = re.search('kernel-(.*).json', connection_file).group(1)
        servers = _list_running_servers()
        for ss in servers:
            response = requests.get(urljoin(ss['url'], 'api/sessions'),
                                    params={'token': ss.get('token', '')})
            for nn in json.loads(response.text):
                if nn['kernel']['id'] == kernel_id:
                    relative_path = nn['notebook']['path']
                    return os.path.join(ss['notebook_dir'], relative_path)

        return ''

    _NOTEBOOK_NAME = _get_notebook_name()
except ImportError:
    _NOTEBOOK_NAME = ''

try:
    if _NOTEBOOK_NAME:
        from tqdm.notebook import tqdm
    else:
        # Either not running notebook or not able to determine
        from tqdm import tqdm
except ImportError:
    tqdm = None

__all__ = ['P_np', 'P_qt', 'abs2', 'all_array_equal', 'dot_HS',
           'get_sample_frequencies', 'hash_array_along_axis', 'mdot',
           'oper_equiv', 'progressbar', 'remove_float_errors', 'tensor',
           'tensor_insert', 'tensor_merge', 'tensor_transpose']

# Pauli matrices
P_qt = [qt.qeye(2),
        qt.sigmax(),
        qt.sigmay(),
        qt.sigmaz()]
P_np = [P.full() for P in P_qt]


def abs2(x: ndarray) -> ndarray:
    r"""
    Fast function to calculate the absolute value squared,

    .. math::

        |\cdot|^2 := \Re(\cdot)^2 + \Im(\cdot)^2

    Equivalent to::

        np.abs(x)**2
    """
    return x.real**2 + x.imag**2


def cexp(x: ndarray) -> ndarray:
    r"""Fast complex exponential.

    Parameters
    ----------
    x : ndarray
        Argument of the complex exponential :math:`\exp(i x)`.

    Returns
    -------
    y : ndarray
        Complex exponential :math:`y = \exp(i x)`.

    References
    ----------
    https://software.intel.com/en-us/forums/intel-distribution-for-python/topic/758148  # noqa
    """
    df_exp = np.empty(x.shape, dtype=np.complex128)
    trig_buf = np.cos(x)
    df_exp.real[:] = trig_buf
    np.sin(x, out=trig_buf)
    df_exp.imag[:] = trig_buf
    return df_exp


def _tensor_product_shape(shape_A: Sequence[int], shape_B: Sequence[int],
                          rank: int):
    """Get shape of the tensor product between A and B of rank rank"""
    broadcast_shape = ()
    # Loop over dimensions from last to first, filling the 'shorter' shape
    # with 1's once it is exhausted
    for dims in zip_longest(shape_A[-rank-1::-1], shape_B[-rank-1::-1],
                            fillvalue=1):
        if 1 in dims:
            # Broadcast 1-d of argument to dimension of other
            broadcast_shape = (max(dims),) + broadcast_shape
        elif len(set(dims)) == 1:
            # Both arguments have same dimension on axis.
            broadcast_shape = dims[:1] + broadcast_shape
        else:
            raise ValueError('Incompatible shapes ' +
                             '{} and {} '.format(shape_A, shape_B) +
                             'for tensor product of rank {}.'.format(rank))

    # Shape of the actual tensor product is product of each dimension,
    # again broadcasting if need be
    tensor_shape = tuple(
        reduce(operator.mul, dimensions) for dimensions in zip_longest(
            shape_A[:-rank-1:-1], shape_B[:-rank-1:-1], fillvalue=1
        )
    )[::-1]

    return broadcast_shape + tensor_shape


def _parse_dims_arg(name: str, dims: Sequence[Sequence[int]],
                    rank: int) -> None:
    """Check if dimension arg for a tensor_* function is correct format"""
    if not len(dims) == rank:
        raise ValueError('{}_dims should be of length '.format(name) +
                         'rank = {}, not {}'.format(rank, len(dims)))

    if not len(set(len(dim) for dim in dims)) == 1:
        # Not all nested lists the same length as required
        raise ValueError('Require all lists in {}_dims '.format(name) +
                         'to be of same length!')


def get_indices_from_identifiers(pulse: 'PulseSequence',
                                 identifiers: Union[None, Sequence[str]],
                                 kind: str) -> Tuple[Sequence[int],
                                                     Sequence[str]]:
    """Get the indices of operators for given identifiers.

    Parameters
    ----------
    pulse : PulseSequence
        The PulseSequence instance for which to get the indices.
    identifiers : sequence of str
        The identifiers whose indices to get.
    kind : str
        Whether to get 'control' or 'noise' operator indices.
    """
    if kind == 'noise':
        pulse_identifiers = pulse.n_oper_identifiers
    elif kind == 'control':
        pulse_identifiers = pulse.c_oper_identifiers

    identifier_to_index_table = {identifier: index for index, identifier in
                                 enumerate(pulse_identifiers)}
    if identifiers is None:
        inds = np.arange(len(pulse_identifiers))
    else:
        try:
            inds = np.array([identifier_to_index_table[identifier]
                             for identifier in identifiers])
        except KeyError:
            raise ValueError('Invalid identifiers given. All available ones ' +
                             'are: {}'.format(pulse_identifiers))

    return inds


def tensor(*args, rank: int = 2,
           optimize: Union[bool, str] = False) -> ndarray:
    """
    Fast, flexible tensor product using einsum. The product is taken over the
    last *rank* axes and broadcast over the remaining axes which thus need to
    follow numpy broadcasting rules. Note that vectors are treated as rank 2
    tensors with shape (1, x) or (x, 1).

    For example, the following shapes are compatible:

     - ``rank == 2`` (e.g. matrices or vectors)::

        (a, b, c, d, d), (a, b, c, e, e) -> (a, b, c, d*e, d*e)
        (a, b, c), (a, d, e) -> (a, b*d, c*e)
        (a, b), (c, d, e) -> (c, a*d, b*e)
        (1, a), (b, 1, c) -> (b, 1, a*c)
     - ``rank == 1``::

        (a, b), (a, c) -> (a, b*c)
        (a, b, 1), (a, c) -> (a, b, c)

    Parameters
    ----------
    args : array_like
        The elements of the tensor product
    rank : int, optional (default: 2)
        The rank of the tensors. E.g., for a Kronecker product between two
        matrices ``rank == 2``. The remaining axes are broadcast over.
    optimize : bool|str, optional (default: False)
        Optimize the tensor contraction order. Passed through to
        :func:`numpy.einsum`.

    Examples
    --------
    >>> Z = np.diag([1, -1])
    >>> np.array_equal(tensor(Z, Z), np.kron(Z, Z))
    True

    >>> A, B = np.arange(2), np.arange(2, 5)
    >>> tensor(A, B, rank=1)
    array([[0, 0, 0, 2, 3, 4]])

    >>> args = np.random.randn(4, 10, 3, 2)
    >>> result = tensor(*args, rank=1)
    >>> result.shape == (10, 3, 2**4)
    True
    >>> result = tensor(*args, rank=2)
    >>> result.shape == (10, 3**4, 2**4)
    True

    >>> A, B = np.random.randn(1, 3), np.random.randn(3, 4)
    >>> result = tensor(A, B)
    >>> result.shape == (1*3, 3*4)
    True

    >>> A, B = np.random.randn(3, 1, 2), np.random.randn(2, 2, 2)
    >>> try:
    ...     result = tensor(A, B, rank=2)
    ... except ValueError as err:  # cannot broadcast over axis 0
    ...     print(err)
    Incompatible shapes (3, 1, 2) and (2, 2, 2) for tensor product of rank 2.
    >>> result = tensor(A, B, rank=3)
    >>> result.shape == (3*2, 1*2, 2*2)
    True

    See Also
    --------
    :func:`numpy.kron`

    :func:`tensor_insert`

    :func:`tensor_merge`

    :func:`tensor_transpose`
    """
    chars = string.ascii_letters
    # All the subscripts we need
    A_chars = chars[:rank]
    B_chars = chars[rank:2*rank]
    subscripts = '...{},...{}->...{}'.format(
        A_chars, B_chars, ''.join(i + j for i, j in zip(A_chars, B_chars))
    )

    def binary_tensor(A, B):
        """Compute the Kronecker product of two tensors"""
        # Add dimensions so that each arg has at least ndim == rank
        while A.ndim < rank:
            A = A[None, :]

        while B.ndim < rank:
            B = B[None, :]

        outshape = _tensor_product_shape(A.shape, B.shape, rank)
        return np.einsum(subscripts, A, B, optimize=optimize).reshape(outshape)

    # Compute the tensor products in a binary tree-like structure, calculating
    # the product of two leaves and working up. This is more memory-efficient
    # than reduce(binary_tensor, args) which computes the products
    # left-to-right.
    n = len(args)
    bit = n % 2
    while n > 1:
        args = args[:bit] + tuple(binary_tensor(*args[i:i+2])
                                  for i in range(bit, n, 2))

        n = len(args)
        bit = n % 2

    return args[0]


def tensor_insert(arr: ndarray, *args, pos: Union[int, Sequence[int]],
                  arr_dims: Sequence[Sequence[int]], rank: int = 2,
                  optimize: Union[bool, str] = False) -> ndarray:
    r"""
    For a tensor product *arr*, insert *args* into the product chain at *pos*.
    E.g, if :math:`\verb|arr|\equiv A\otimes B\otimes C` and
    :math:`\verb|pos|\equiv 2`, the result will be the tensor product

    .. math::
        A\otimes B\otimes\left[\bigotimes_{X\in\verb|args|}X\right]\otimes C.

    This function works in a similar way to :func:`numpy.insert` and the
    following would be functionally equivalent in the case that the constituent
    tensors of the product *arr* are known:

    >>> tensor_insert(tensor(*arrs, rank=rank), *args, pos=pos, arr_dims=...,
    ...               rank=rank)

    >>> tensor(*np.insert(arrs, pos, args, axis=0), rank=rank)


    Parameters
    ----------
    arr : ndarray
        The tensor product in whose chain the other args should be inserted
    *args : ndarray
        The tensors to be inserted in the product chain
    pos : int|sequence of ints
        The position(s) at which the args are inserted in the product chain. If
        an int and ``len(args) > 1``, it is repeated so that all args are
        inserted in a row. If a sequence, it should indicate the indices in the
        original tensor product chain that led to *arr* before which *args*
        should be inserted.
    arr_dims : array_like, shape (rank, n_const)
        The last *rank* dimensions of the *n_const* constituent tensors of the
        tensor product *arr* as a list of lists with the list at position *i*
        containing the *i*-th relevant dimension of all args. Since the remaing
        axes are broadcast over, their shape is irrelevant.

        For example, if ``arr = tensor(a, b, c, rank=2)`` and ``a,b,c`` have
        shapes ``(2, 3, 4), (5, 2, 2, 1), (2, 2)``,
        ``arr_dims = [[3, 2, 2], [4, 1, 2]]``.
    rank : int, optional (default: 2)
        The rank of the tensors. E.g., for a Kronecker product between two
        vectors, ``rank == 1``, and between two matrices ``rank == 2``. The
        remaining axes are broadcast over.
    optimize : bool|str, optional (default: False)
        Optimize the tensor contraction order. Passed through to
        :func:`numpy.einsum`.


    Examples
    --------
    >>> I, X, Y, Z = P_np
    >>> arr = tensor(X, I)
    >>> r = tensor_insert(arr, Y, Z, arr_dims=[[2, 2], [2, 2]], pos=0)
    >>> np.allclose(r, tensor(Y, Z, X, I))
    True
    >>> r = tensor_insert(arr, Y, Z, arr_dims=[[2, 2], [2, 2]], pos=1)
    >>> np.allclose(r, tensor(X, Y, Z, I))
    True
    >>> r = tensor_insert(arr, Y, Z, arr_dims=[[2, 2], [2, 2]], pos=2)
    >>> np.allclose(r, tensor(X, I, Y, Z))
    True

    Other ranks and different dimensions:

    >>> from numpy.random import randn
    >>> A, B, C = randn(2, 3, 1, 2), randn(2, 2, 2, 2), randn(3, 2, 1)
    >>> arr = tensor(A, C, rank=3)
    >>> r = tensor_insert(arr, B, pos=1, rank=3,
    ...                   arr_dims=[[3, 3], [1, 2], [2, 1]])
    >>> np.allclose(r, tensor(A, B, C, rank=3))
    True

    >>> arrs, args = randn(2, 2, 2), randn(2, 2, 2)
    >>> arr_dims = [[2, 2], [2, 2]]
    >>> r = tensor_insert(tensor(*arrs), *args, pos=(0, 1), arr_dims=arr_dims)
    >>> np.allclose(r, tensor(args[0], arrs[0], args[1], arrs[1]))
    True
    >>> r = tensor_insert(tensor(*arrs), *args, pos=(0, 0), arr_dims=arr_dims)
    >>> np.allclose(r, tensor(*args, *arrs))
    True
    >>> r = tensor_insert(tensor(*arrs), *args, pos=(1, 2), arr_dims=arr_dims)
    >>> np.allclose(r, tensor(*np.insert(arrs, (1, 2), args, axis=0)))
    True

    See Also
    --------
    :func:`tensor`

    :func:`tensor_merge`

    :func:`tensor_transpose`

    :func:`numpy.kron`

    :func:`numpy.insert`
    """
    if len(args) == 0:
        raise ValueError('Require nonzero number of args!')

    if isinstance(pos, int):
        pos = (pos,)
        if len(args) > 1:
            # Inserting all args at same position, perform their tensor product
            # using tensor and insert the result instead of iteratively insert
            # one by one
            args = (tensor(*args, rank=rank, optimize=optimize),)
    else:
        if not len(pos) == len(args):
            raise ValueError('Expected pos to be either an int or a ' +
                             'sequence of the same length as the number of ' +
                             'args, not length {}'.format(len(pos)))

    _parse_dims_arg('arr', arr_dims, rank)

    def _tensor_insert_subscripts(ndim, pos, rank):
        """Get einsum string for the contraction"""
        ins_chars = string.ascii_letters[:rank]
        arr_chars = string.ascii_letters[rank:(ndim+1)*rank]
        subscripts = '...{},...{}->...{}'.format(
            ins_chars, arr_chars,
            arr_chars[:pos] + ''.join(
                ins_chars[i] + arr_chars[pos+i*ndim:pos+(i+1)*ndim]
                for i in range(rank)
            )
        )

        return subscripts

    def single_tensor_insert(arr, ins, arr_dims, pos):
        """Insert a single tensor *ins* into *arr* at position *pos*."""
        subscripts = _tensor_insert_subscripts(len(arr_dims[0]), pos, rank)
        outshape = _tensor_product_shape(ins.shape, arr.shape, rank)
        # Need to reshape arr to the rank*ndim-dimensional shape that's the
        # output of the regular tensor einsum call
        flat_arr_dims = [dim for axis in arr_dims for dim in axis]
        reshaped_arr = arr.reshape(*arr.shape[:-rank], *flat_arr_dims)
        result = np.einsum(subscripts, ins, reshaped_arr,
                           optimize=optimize).reshape(outshape)

        return result

    # Insert args one after the other, starting at lowest index
    result = arr.copy()
    # Make a deep copy of arr_dims and pos as we modify them
    carr_dims = [list(axis[:]) for axis in arr_dims]
    cpos = list(pos).copy()
    # Number of constituent tensors of the tensor product arr
    ndim = len(arr_dims[0])
    divs, pos = zip(*[divmod(p, ndim) if p != ndim else (0, p) for p in pos])
    for i, (p, div, arg, arg_counter) in enumerate(sorted(
            zip_longest(pos, divs, args, range(len(args)), fillvalue=pos[0]),
            key=operator.itemgetter(0))):

        if div not in (-1, 0):
            raise IndexError('Invalid position {} '.format(cpos[i]) +
                             'specified. Must be between ' +
                             '-{0} and {0}.'.format(ndim))

        # Insert argument arg at position p+i (since every iteration the index
        # shifts by 1)
        try:
            result = single_tensor_insert(result, arg, carr_dims, p+i)
        except ValueError as err:
            raise ValueError(
                'Could not insert arg {} with shape '.format(arg_counter) +
                '{} into the array with shape '.format(result.shape) +
                '{} at position {}.'.format(arg.shape, p)
            ) from err

        # Update arr_dims
        for axis, d in zip(carr_dims, arg.shape[-rank:]):
            axis.insert(p, d)

    return result


def tensor_merge(arr: ndarray, ins: ndarray, pos: Sequence[int],
                 arr_dims: Sequence[Sequence[int]],
                 ins_dims: Sequence[Sequence[int]],
                 rank: int = 2, optimize: Union[bool, str] = False) -> ndarray:
    r"""
    For two tensor products *arr* and *ins*, merge *ins* into the product chain
    at indices *pos*. E.g, if :math:`\verb|arr|\equiv A\otimes B\otimes C`,
    :math:`\verb|ins|\equiv D\otimes E`, and :math:`\verb|pos|\equiv [1, 2]`,
    the result will be the tensor product

    .. math::
        A\otimes D\otimes B\otimes E\otimes C.

    This function works in a similar way to :func:`numpy.insert` and
    :func:`tensor_insert`.

    Parameters
    ----------
    arr : ndarray
        The tensor product in whose chain the other args should be inserted
    ins : ndarray
        The tensor product to be inserted in the product chain
    pos : sequence of ints
        The positions at which the constituent tensors of *ins* are inserted in
        the product chain. Should indicate the indices in the original tensor
        product chain that led to *arr* before which the constituents of *ins*
        should be inserted.
    arr_dims : array_like, shape (rank, n_const)
        The last *rank* dimensions of the *n_const* constituent tensors of the
        tensor product *arr* as a list of lists with the list at position *i*
        containing the *i*-th relevant dimension of all args. Since the remaing
        axes are broadcast over, their shape is irrelevant.

        For example, if ``arr = tensor(a, b, c, rank=2)`` and ``a,b,c`` have
        shapes ``(2, 3, 4), (5, 2, 2, 1), (2, 2)``,
        ``arr_dims = [[3, 2, 2], [4, 1, 2]]``.
    ins_dims : array_like, shape (rank, n_const)
        The last *rank* dimensions of the *n_const* constituent tensors of the
        tensor product *ins* as a list of lists with the list at position *i*
        containing the *i*-th relevant dimension of *ins*. Since the remaing
        axes are broadcast over, their shape is irrelevant.
    rank : int, optional (default: 2)
        The rank of the tensors. E.g., for a Kronecker product between two
        vectors, ``rank == 1``, and between two matrices ``rank == 2``. The
        remaining axes are broadcast over.
    optimize : bool|str, optional (default: False)
        Optimize the tensor contraction order. Passed through to
        :func:`numpy.einsum`.

    Examples
    --------
    >>> I, X, Y, Z = P_np
    >>> arr = tensor(X, Y, Z)
    >>> ins = tensor(I, I)
    >>> r1 = tensor_merge(arr, ins, pos=[1, 2], arr_dims=[[2]*3, [2]*3],
    ...                   ins_dims=[[2]*2, [2]*2])
    >>> np.allclose(r1, tensor(X, I, Y, I, Z))
    True
    >>> r2 = tensor_merge(ins, arr, pos=[0, 1, 2], arr_dims=[[2]*2, [2]*2],
    ...                   ins_dims=[[2]*3, [2]*3])
    >>> np.allclose(r1, r2)
    True

    :func:`tensor_insert` can provide the same functionality in some cases:

    >>> arr = tensor(Y, Z)
    >>> ins = tensor(I, X)
    >>> r1 = tensor_merge(arr, ins, pos=[0, 0], arr_dims=[[2]*2, [2]*2],
    ...                   ins_dims=[[2]*2, [2]*2])
    >>> r2 = tensor_insert(arr, I, X, pos=[0, 0], arr_dims=[[2]*2, [2]*2])
    >>> np.allclose(r1, r2)
    True

    Also tensors of rank other than 2 and numpy broadcasting are supported:

    >>> arr = np.random.randn(2, 10, 3, 4)
    >>> ins = np.random.randn(2, 10, 3, 2)
    >>> r = tensor_merge(tensor(*arr, rank=1), tensor(*ins, rank=1), [0, 1],
    ...                  arr_dims=[[4, 4]], ins_dims=[[2, 2]], rank=1)
    >>> np.allclose(r, tensor(ins[0], arr[0], ins[1], arr[1], rank=1))
    True

    See Also
    --------
    :func:`tensor`

    :func:`tensor_insert`

    :func:`tensor_transpose`

    :func:`numpy.kron`

    :func:`numpy.insert`
    """
    # Parse dimension args
    for arg_name, arg_dims in zip(('arr', 'ins'), (arr_dims, ins_dims)):
        _parse_dims_arg(arg_name, arg_dims, rank)

    ins_ndim = len(ins_dims[0])
    arr_ndim = len(arr_dims[0])
    ins_chars = string.ascii_letters[:ins_ndim*rank]
    arr_chars = string.ascii_letters[ins_ndim*rank:(ins_ndim+arr_ndim)*rank]
    out_chars = ''
    for r in range(rank):
        arr_part = arr_chars[r*arr_ndim:(r+1)*arr_ndim]
        ins_part = ins_chars[r*ins_ndim:(r+1)*ins_ndim]
        for i, (p, ins_p) in enumerate(sorted(zip(pos, ins_part))):
            if p != arr_ndim:
                div, p = divmod(p, arr_ndim)
                if div not in (-1, 0):
                    raise IndexError('Invalid position {} '.format(pos[i]) +
                                     'specified. Must be between ' +
                                     '-{0} and {0}.'.format(arr_ndim))
            arr_part = arr_part[:p+i] + ins_p + arr_part[p+i:]

        out_chars += arr_part

    subscripts = '...{},...{}->...{}'.format(
        ins_chars, arr_chars, out_chars
    )

    outshape = _tensor_product_shape(ins.shape, arr.shape, rank)
    # Need to reshape arr to the rank*ndim-dimensional shape that's the
    # output of the regular tensor einsum call
    flat_arr_dims = [dim for axis in arr_dims for dim in axis]
    flat_ins_dims = [dim for axis in ins_dims for dim in axis]

    # Catch exceptions from wrong ins/arr_dims arguments
    try:
        ins_reshaped = ins.reshape(*ins.shape[:-rank], *flat_ins_dims)
    except ValueError as err:
        raise ValueError('ins_dims not compatible with ins.shape[-rank:] = ' +
                         '{}'.format(ins.shape[-rank:])) from err
    try:
        arr_reshaped = arr.reshape(*arr.shape[:-rank], *flat_arr_dims)
    except ValueError as err:
        raise ValueError('arr_dims not compatible with arr.shape[-rank:] = ' +
                         '{}'.format(arr.shape[-rank:])) from err

    result = np.einsum(subscripts, ins_reshaped, arr_reshaped,
                       optimize=optimize).reshape(outshape)

    return result


def tensor_transpose(arr: ndarray, order: Sequence[int],
                     arr_dims: Sequence[Sequence[int]],
                     rank: int = 2) -> ndarray:
    r"""
    Transpose the order of a tensor product chain.

    Parameters
    ----------
    arr : ndarray
        The tensor product whose chain should be reordered.
    order : sequence of ints
        The transposition order. If ``arr == tensor(A, B)`` and
        ``order == (1, 0)``, the result will be ``tensor(B, A)``.
    arr_dims : array_like, shape (rank, n_const)
        The last *rank* dimensions of the *n_const* constituent tensors of the
        tensor product *arr* as a list of lists with the list at position *i*
        containing the *i*-th relevant dimension of all args. Since the remaing
        axes are broadcast over, their shape is irrelevant.

        For example, if ``arr = tensor(a, b, c, rank=2)`` and ``a,b,c`` have
        shapes ``(2, 3, 4), (5, 2, 2, 1), (2, 2)``,
        ``arr_dims = [[3, 2, 2], [4, 1, 2]]``.
    rank : int, optional (default: 2)
        The rank of the tensors. E.g., for a Kronecker product between two
        vectors, ``rank == 1``, and between two matrices ``rank == 2``. The
        remaining axes are broadcast over.

    Returns
    -------
    transposed_arr : ndarray
        The tensor product *arr* with its order transposed according to *order*

    Examples
    --------
    >>> I, X, Y, Z = P_np
    >>> arr = tensor(X, Y, Z)
    >>> transposed = tensor_transpose(arr, [1, 2, 0], arr_dims=[[2, 2, 2]]*2)
    >>> np.allclose(transposed, tensor(Y, Z, X))
    True

    See Also
    --------
    :func:`tensor`

    :func:`tensor_insert`

    :func:`tensor_transpose`

    :func:`numpy.kron`
    """
    _parse_dims_arg('arr', arr_dims, rank)

    ndim = len(arr_dims[0])
    # Number of axes that are broadcast over
    n_broadcast = len(arr.shape[:-rank])
    transpose_axes = [i for i in range(n_broadcast)] + \
        [n_broadcast + r*ndim + o for r in range(rank) for o in order]

    # Need to reshape arr to the rank*ndim-dimensional shape that's the
    # output of the regular tensor einsum call
    flat_arr_dims = [dim for axis in arr_dims for dim in axis]

    # Catch exceptions from wrong arr_dims argument
    try:
        arr_reshaped = arr.reshape(*arr.shape[:-rank], *flat_arr_dims)
    except ValueError as err:
        raise ValueError('arr_dims not compatible with arr.shape[-rank:] = ' +
                         '{}'.format(arr.shape[-rank:])) from err

    try:
        result = arr_reshaped.transpose(*transpose_axes).reshape(arr.shape)
    except TypeError as type_err:
        raise TypeError("Could not transpose the order. Are all elements of " +
                        "'order' ints?") from type_err
    except ValueError as val_err:
        raise ValueError("Could not transpose the order. Are all elements " +
                         "of 'order' unique and match the array?") from val_err

    return result


def mdot(arr: Sequence, axis: int = 0) -> ndarray:
    """Multiple matrix products along axis"""
    return reduce(np.matmul, np.swapaxes(arr, 0, axis))


def remove_float_errors(arr: ndarray, eps_scale: float = None):
    """
    Clean up arr by removing floating point numbers smaller than the dtype's
    precision multiplied by eps_scale. Treats real and imaginary parts
    separately.
    """
    if eps_scale is None:
        atol = np.finfo(arr.dtype).eps*arr.shape[-1]
    else:
        atol = np.finfo(arr.dtype).eps*eps_scale

    # Hack around arr.imag sometimes not being writable
    if arr.dtype == complex:
        arr = arr.real + 1j*arr.imag
        arr.real[np.abs(arr.real) <= atol] = 0
        arr.imag[np.abs(arr.imag) <= atol] = 0
    else:
        arr = arr.real
        arr.real[np.abs(arr.real) <= atol] = 0

    return arr


def oper_equiv(psi: Union[Operator, State],
               phi: Union[Operator, State],
               eps: float = None,
               normalized: bool = False) -> Tuple[bool, float]:
    r"""
    Checks whether psi and phi are equal up to a global phase, i.e.

    .. math::
        |\psi\rangle = e^{i\chi}|\phi\rangle \Leftrightarrow
        \langle \phi|\psi\rangle = e^{i\chi},

    and returns the phase. If the first return value is false, the second is
    meaningless in this context. psi and phi can also be operators.

    Parameters
    ----------
    psi, phi : Qobj or array_like
        Vectors or operators to be compared
    eps : float
        The tolerance below which the two objects are treated as equal, i.e.,
        the function returns ``True`` if ``abs(1 - modulus) <= eps``.
    normalized : bool
        Flag indicating if *psi* and *phi* are normalized with respect to the
        Hilbert-Schmidt inner product :func:`dot_HS`.

    Examples
    --------
    >>> psi = qt.sigmax()
    >>> phi = qt.sigmax()*np.exp(1j*1.2345)
    >>> oper_equiv(psi, phi)
    (True, 1.2345)
    """
    psi, phi = [obj.full() if isinstance(obj, qt.Qobj) else obj
                for obj in (psi, phi)]

    if eps is None:
        # Tolerance the floating point eps times the # of flops for the matrix
        # multiplication, i.e. for psi and phi n x m matrices 2*n**2*m
        eps = max(np.finfo(psi.dtype).eps, np.finfo(phi.dtype).eps) *\
            np.prod(psi.shape)*phi.shape[-1]*2
        if not normalized:
            # normalization introduces more floating point error
            eps *= (np.prod(psi.shape)*phi.shape[-1]*2)**2

    inner_product = (psi.T.conj() @ phi).trace()
    if normalized:
        norm = 1
    else:
        norm = linalg.norm(psi)*linalg.norm(phi)

    phase = np.angle(inner_product)
    modulus = abs(inner_product)

    return abs(norm - modulus) <= eps, phase


def dot_HS(U: Operator, V: Operator, eps: float = None) -> float:
    r"""Return the Hilbert-Schmidt inner product of U and V,

    .. math::
        \langle U, V\rangle_\mathrm{HS} := \mathrm{tr}(U^\dagger V).

    Parameters
    ----------
    U, V : Qobj or ndarray
        Objects to compute the inner product of.

    Returns
    -------
    result : float, complex
        The result rounded to precision eps.

    Examples
    --------
    >>> U, V = qt.sigmax(), qt.sigmay()
    >>> dot_HS(U, V)
    0.0
    >>> dot_HS(U, U)
    2.0
    """
    if isinstance(U, qt.Qobj):
        U = U.full()
    if isinstance(V, qt.Qobj):
        V = V.full()

    if eps is None:
        # Tolerance is the dtype precision times the number of flops for the
        # matrix multiplication times two to be on the safe side
        try:
            eps = np.finfo(U.dtype).eps*np.prod(U.shape)*V.shape[-1]*2
        except ValueError:
            # dtype is int and therefore exact
            eps = 0

    if eps == 0:
        decimals = 0
    else:
        decimals = abs(int(np.log10(eps)))

    res = np.round(np.einsum('...ij,...ij', U.conj(), V), decimals)
    return res if res.imag.any() else res.real


def get_sample_frequencies(pulse: 'PulseSequence', n_samples: int = 200,
                           spacing: str = 'log',
                           symmetric: bool = True) -> ndarray:
    """
    Get *n_samples* sample frequencies spaced either 'linear' or 'log'
    symmetrically around zero.

    The ultraviolet cutoff is taken to be two orders of magnitude larger than
    the timescale of the pulse tau. In the case of log spacing, the values are
    clipped in the infrared at two orders of magnitude below the timescale of
    the pulse.

    Parameters
    ----------
    pulse : PulseSequence
        The pulse to get frequencies for.
    n_samples : int, optional
        The number of frequency samples. Default is 200.
    spacing : str, optional
        The spacing of the frequencies. Either 'log' or 'linear', default is
        'log'.
    symmetric : bool, optional
        Whether the frequencies should be symmetric around zero or positive
        only. Default is True.

    Returns
    -------
    omega : ndarray
        The frequencies.
    """
    tau = pulse.t[-1]
    if spacing == 'linear':
        if symmetric:
            freqs = np.linspace(-1e2/tau, 1e2/tau, n_samples)*2*np.pi
        else:
            freqs = np.linspace(0, 1e2/tau, n_samples)*2*np.pi
    elif spacing == 'log':
        if symmetric:
            freqs = np.geomspace(1e-2/tau, 1e2/tau, n_samples//2)*2*np.pi
            freqs = np.concatenate([-freqs[::-1], freqs])
        else:
            freqs = np.geomspace(1e-2/tau, 1e2/tau, n_samples)*2*np.pi
    else:
        raise ValueError("spacing should be either 'linear' or 'log'.")

    return freqs


def symmetrize_spectrum(S: ndarray, omega: ndarray) -> Tuple[ndarray, ndarray]:
    r"""
    Symmetrize a one-sided power spectrum around zero frequency.

    Parameters
    ----------
    S : ndarray, shape (..., n_omega)
        The one-sided power spectrum.
    omega : ndarray, shape (n_omega,)
        The positive and strictly increasing frequencies.

    Returns
    -------
    S : ndarray, shape (..., 2*n_omega)
        The two-sided power spectrum.
    omega : ndarray, shape (2*n_omega,)
        The frequencies mirrored about zero.

    Notes
    -----
    The two-sided power spectral density is in the symmetric case given by
    :math:`S^{(1)}(\omega) = 2S^{(2)}(\omega)`.
    """
    # Catch zero frequency component
    if omega[0] == 0:
        ix = 1
    else:
        ix = 0

    omega = np.concatenate((-omega[::-1], omega[ix:]))
    S = np.concatenate((S[..., ::-1], S[ix:]), axis=-1)/2
    return S, omega


def hash_array_along_axis(arr: ndarray, axis: int = 0) -> List[int]:
    """Return the hashes of arr along the first axis"""
    return [hash(arr.tobytes()) for arr in np.swapaxes(arr, 0, axis)]


def all_array_equal(it: Iterable) -> bool:
    """
    Return ``True`` if all array elements of ``it`` are equal by hashing the
    bytes representation of each array. Note that this is not thread-proof.
    """
    return len(set(hash(i.tobytes()) for i in it)) == 1


def _simple_progressbar(iterable: Iterable, desc: str = "Computing",
                        size: int = 25, count: int = None, file=None):
    """https://stackoverflow.com/a/34482761"""
    if count is None:
        try:
            count = len(iterable)
        except TypeError:
            raise TypeError("Require total number of iterations 'count'.")

    file = sys.stdout if file is None else file

    if desc:
        # tqdm desc compatibility
        desc = desc.strip(': ') + ': '

    def show(j):
        x = int(size*j/count)
        file.write("\r{}[{}{}] {} %".format(desc, "#"*x, "."*(size - x),
                                            int(100*j/count)))
        file.flush()

    show(0)
    for i, item in enumerate(iterable):
        yield item
        show(i + 1)

    file.write("\n")
    file.flush()


def progressbar(iterable: Iterable, *args, **kwargs):
    """
    Progress bar for loops. Uses tqdm if available or a quick-and-dirty
    implementation from stackoverflow.

    Usage::

        for i in progressbar(range(10)):
            do_something()
    """
    if tqdm is not None:
        return tqdm(iterable, *args, **kwargs)

    return _simple_progressbar(iterable, *args, **kwargs)


def progressbar_range(*args, show_progressbar: Optional[bool] = True,
                      **kwargs):
    """Wrapper for range() that shows a progressbar dependent on a kwarg.

    Parameters
    ----------
    *args :
        Positional arguments passed through to :func:`range`.
    show_progressbar : bool, optional
        Return a range iterator with or without a progressbar.
    **kwargs :
        Keyword arguments passed through to :func:`progressbar`.

    Returns
    -------
    it : Iterator
        Range iterator dressed with a progressbar if ``show_progressbar=True``.
    """
    if show_progressbar:
        return progressbar(range(*args), **kwargs)

    return range(*args)


class CalculationError(Exception):
    """Indicates a quantity could not be computed."""

    def __init__(self, message: str) -> None:
        super().__init__(message)



# -*- coding: utf-8 -*-
import os
import re
import sys

from setuptools import setup


def read(*args):
    return open(os.path.join(os.path.dirname(__file__), *args),
                encoding='utf8').read()


def extract_version(version_file):
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)

    raise RuntimeError("Unable to find version string.")


if sys.version_info < (3, 5):
    sys.stderr.write('ERROR: You need Python 3.5 or later '
                     'to install this package.\n')
    exit(1)

setup(name='filter_functions',
      version=extract_version(read('filter_functions', '__init__.py')),
      description='Package for efficient calculation of generalized filter functions',
      long_description=read('README.md'),
      long_description_content_type='text/markdown',
      url='https://github.com/qutech/filter_functions',
      author='Quantum Technology Group, RWTH Aachen University',
      author_email='tobias.hangleiter@rwth-aachen.de',
      packages=['filter_functions'],
      package_dir={'filter_functions': 'filter_functions'},
      install_requires=['numpy', 'scipy', 'matplotlib', 'qutip', 'opt_einsum',
                        'sparse'],
      extras_require={
          'fancy_progressbar': ['tqdm', 'requests'],
          'doc': ['ipython', 'ipykernel', 'nbsphinx', 'numpydoc', 'sphinx',
                  'jupyter_client', 'sphinx_rtd_theme'],
          'tests': ['pytest', 'coverage', 'coveralls'],
      },
      test_suite='tests',
      classifiers=[
          'Programming Language :: Python :: 3',
          'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
          'Operating System :: OS Independent',
          'Topic :: Scientific/Engineering :: Physics',
      ])
